%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  rpo.pl   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% USES: a_intrpo, list, fix_rpo, precstat, int_rpo, (ordre:is_variable)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% try_rpo(Rules,Prec,Stat,PrecOut,StatOut)
% check_rules_rpo(Rules,Prec,Stat,RulesOut)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

try_rpo(Rules,Prec,Stat,PrecOut,StatOut):-
	sort(Rules,Rules0),
	apply_simplifications(Rules0,Prec,Prec1,Stat,Stat1,Rules1),
	step_rpo(Rules1,Prec1,Stat1,PrecOut,StatOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% elimina regles per >rpo1 (ja conegut)
% aplicaci de rpo per > i = coneguts -> generaci de constr. de subt.
% No backtracking.

apply_simplifications([],Prec,Prec,Stat,Stat,[]):-!.          
apply_simplifications(Rules,Prec,PrecOut,Stat,StatOut,RulesOut):-
	check_rules_rpo(Rules,Prec,Stat,Rules1),
	simplify_applic_rpo(Rules1,Prec,Stat,Rules2,Prec2,Stat2),
	end_apply_simplifications(Prec,Prec2,Rules1,Rules2,Stat2,PrecOut,StatOut,RulesOut),!.

end_apply_simplifications(Prec,Prec1,Rules,Rules1,Stat1,Prec11,Stat1,Rules11):-
	sort(Rules,Rules00),sort(Rules1,Rules11), Rules00=Rules11,
	sort(Prec,Prec00),sort(Prec1,Prec11), Prec00=Prec11,!.
end_apply_simplifications(_,Prec1,_,Rules1,Stat1,PrecOut,StatOut,RulesOut):-
	apply_simplifications(Rules1,Prec1,PrecOut,Stat1,StatOut,RulesOut),!.

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% aplicaci de rpo per > i = coneguts -> generaci de constr. de subt.
% Si examinant variables no es possible usar subterm -> es genera
% precedncia. 

% No backtracking.


simplify_applic_rpo(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	simplify_applic_rpo(Rules,Prec,Stat,[],RulesOut,PrecOut,StatOut),!.

	
simplify_applic_rpo([],Prec,Stat,Rules0,Rules0,Prec,Stat):-!.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..['='|[L,R]],!,
	L=..[F|ArgsL],
	R=..[G|ArgsR],
	fer_iguals_tops(F,G,Prec,Stat,Prec1,Stat1),
	quin_stat(F,Stat1,StatF0),
	aval_stat(StatF0,ArgsL,ArgsR,F,Prec1,Stat1,StatF,Stat2,'='),
	simpl_args(StatF,Prec1,Stat2,ArgsL,ArgsR,ArgsL1,ArgsR1),!,
	tractar_args_restants(StatF,Prec1,Stat2,L,ArgsL1,R,ArgsR1,'=',Rules0,Rules1),
	simplify_applic_rpo(Rules,Prec1,Stat2,Rules1,RulesOut,PrecOut,StatOut),!.
%a partir d'aqu regles > o >=.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[_|[L,R]],
	L=..[F|_],
	R=..[G|ArgsR],
	member212(F>G, Prec),
	gen_rules_rec_call(>,L,ArgsR,Rules0,Rules1),!,
	simplify_applic_rpo(Rules,Prec,Stat,Rules1,RulesOut,PrecOut,StatOut),!.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[Op|[L,R]],
	L=..[F|ArgsL],
	R=..[G|ArgsR],
	igual(F,G,Prec),!,
	quin_stat(F,Stat,StatF0),
	aval_stat(StatF0,ArgsL,ArgsR,F,Prec,Stat,StatF,Stat1,Op),
	simpl_args(StatF,Prec,Stat1,ArgsL,ArgsR,ArgsL1,ArgsR1),!,
	tractar_args_restants(StatF,Prec,Stat1,L,ArgsL1,R,ArgsR1,Op,Rules0,Rules1),
	simplify_applic_rpo(Rules,Prec,Stat1,Rules1,RulesOut,PrecOut,StatOut),!.
% si L>R i F>=G comprovem si F=G seria possible. Si ho es deixem la regla
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[>|[L,R]],
	L=..[F|ArgsL],
	R=..[G|ArgsR],
	member212(F>=G, Prec),
	quin_statFG(F,G,Stat,StatFG),
	simpl_args(StatFG,Prec,Stat,ArgsL,ArgsR,DLR,_),DLR=[_|_],!,
	simplify_applic_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
% si l'anterior ha fallat segur que cal F>G.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[>|[L,R]],
	L=..[F|_],
	R=..[G|_],
	member212(F>=G, Prec),!,
	generarF(F>G,Prec,Prec1,Stat,Stat1),
	simplify_applic_rpo(Rules,Prec1,Stat1,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[Op|[L,R]],
	functor(L,F,0),!,
	generar_prec1(F,Op,R,Prec,Prec1,Stat,Stat1),  %segur que R ground: ha passat check
	simplify_applic_rpo(Rules,Prec1,Stat1,Rules0,RulesOut,PrecOut,StatOut),!.
% si g>f assegurar subterm_possible
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[_|[L,R]],
	L=..[F|Args],R=..[G|_],
	member212(G>F,Prec),!,
	subterm_possible(Args,R,Prec,Stat),!,
	simplify_applic_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[_|[L,R]],
	L=..[F|_],R=..[G|_],
	exist_prec_g_ge_eq(F,G,Prec),!,
	simplify_applic_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
% o no hi ha relaci fixada entre F i G, o be tenim G>=F
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[_|[L,R]], L=..[F|ArgsL], R=..[G|_],
	\+(is_variable(R)), % innecessari. Segur que no ho s 
	all_variables(ArgsL),!,
	generarF(F>=G,Prec,Prec1,Stat,Stat1),!,
	simplify_applic_rpo(Rules,Prec1,Stat1,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[_|[L,R]], L=..[_|Args],
	subterm_possible(Args,R,Prec,Stat),!,
	simplify_applic_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
% o no hi ha relaci fixada entre F i G, o be tenim G>=F i no subterm_possible
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	Rule=..[_|[L,R]], L=..[F|_], R=..[G|_],
	generarF(F>=G,Prec,Prec1,Stat,Stat1),!,
	simplify_applic_rpo(Rules,Prec1,Stat1,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.
%aqu mai arriba (deixo per comprovar)
simplify_applic_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut,PrecOut,StatOut):-
	simplify_applic_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut,PrecOut,StatOut),!.


quin_statFG(F,G,Stat,StatG0):-
	quin_stat(F,Stat,cap),!,
	quin_stat(G,Stat,StatG0),!.
quin_statFG(F,G,Stat,StatF0):-
	quin_stat(F,Stat,StatF0),!,
	quin_stat(G,Stat,StatG0),!,
	(StatG0=cap ; StatG0=StatF0).

%%%%%
gen_rules_rec_call(_,_,[],Rules,Rules):-!.
gen_rules_rec_call(Op,L,[R|Terms],Rules,RulesOut):-
	NewRule=..[Op|[L,R]],
	gen_rules_rec_call(Op,L,Terms,[NewRule|Rules],RulesOut),!.

%%%%%
tractar_args_restants(_,_,_,_,[],_,L,'=',Rules,Rules):-!,L=[].
tractar_args_restants(_,_,_,_,L,_,[],'=',Rules,Rules):-!,L=[].
tractar_args_restants(_,_,_,_,[],_,_,'>',_,_):-!,fail.
tractar_args_restants(_,_,_,_,_,_,[],_,Rules,Rules):-!.
tractar_args_restants(_,_,_,_,[ArgL],_,[ArgR],Op,Rules0,[New_Rule|Rules0]):-!,
	New_Rule=..[Op|[ArgL,ArgR]],!.
tractar_args_restants(cap,_,_,L,_,R,_,Op,Rules0,[Same_Rule|Rules0]):-!,
	Same_Rule=..[Op|[L,R]],!.
% a partir d'aqu mul o lex o lexr
% = i mul
tractar_args_restants(mul,Prec,Stat,L,ArgsL,R,ArgsR,'=',Rules0,RulesOut):-!,
	functor(L,F,ArL),
	functor(R,G,ArR),!, ArL=ArR,
	tractar_args_mul(Prec,Stat,F,G,ArgsL,ArgsR,'=',Rules0,RulesOut,_),!.
% = i lex o lexr
tractar_args_restants(_,Prec,Stat,_,ArgsL,_,ArgsR,'=',Rules0,RulesOut):-!,
	gen_ig_args(ArgsL,ArgsR,Prec,Stat,Rules0,RulesOut),!.
%a partir d'aqu regles > o >= amb mul,lex o lexr
tractar_args_restants(lex,_,_,L,[ArgL],_,[ArgR|ArgsR],_,Rules0,RulesOut):-!,
	gen_rules_rec_call(>,L,ArgsR,[ArgL>ArgR|Rules0],RulesOut),!.
tractar_args_restants(lex,Prec,Stat,L,[A|_],_,[X|ArgsR],_,Rules0,RulesOut):-
	segur_no_iguals(A,X,Prec,Stat),!,
	gen_rules_rec_call(>,L,ArgsR,[A>X|Rules0],RulesOut),!.
tractar_args_restants(lex,_,_,L,_,R,_,Op,Rules0,[New_Rule|Rules0]):-
	New_Rule=..[Op|[L,R]],!.
tractar_args_restants(lexr,Prec,Stat,L,ArgsL,R,ArgsR,Op,Rules0,RulesOut):-!,
	reverse1(ArgsL,ArgsL1),
	reverse1(ArgsR,ArgsR1),
	tractar_args_restants(lex,Prec,Stat,L,ArgsL1,R,ArgsR1,Op,Rules0,RulesOut),!.
tractar_args_restants(mul,Prec,Stat,L,ArgsL,R,ArgsR,Op,Rules0,RulesOut):-!,
	simp_subt_gr(ArgsL,ArgsR,Prec,Stat,ArgsR1,GR),
	((GR=yes,Op1='>=');(GR=no,Op1=Op)),
	functor(L,F,_),functor(R,G,_),
	tractar_args_mul(Prec,Stat,F,G,ArgsL,ArgsR1,Op1,Rules0,Rules1,Que),
	((Que=ok, RulesOut=Rules1); 
	 (Que=no, Rule=..[Op|[L,R]],RulesOut=[Rule|Rules1])),!.

tractar_args_mul(_,_,_,_,L,[],'=',Rules,Rules,ok):-!,L=[].
tractar_args_mul(_,_,_,_,[],L,'=',Rules,Rules,ok):-!,L=[].
tractar_args_mul(_,_,_,_,[],_,'>',_,_,_):-!,fail.
tractar_args_mul(_,_,_,_,_,[],_,Rules,Rules,ok):-!.
tractar_args_mul(_,_,_,_,[ArgL],[ArgR],Op,Rules0,[NewRule|Rules0],ok):-!,
	NewRule=..[Op|[ArgL,ArgR]],!.
tractar_args_mul(_,_,_,_,[ArgL],ArgsR,Op,Rules0,RulesOut,ok):-
	(Op='>';Op='>='),!,
	gen_rules_rec_call(>,ArgL,ArgsR,Rules0,RulesOut),!.
%sabem segur ArgsR=[_,_|_]
tractar_args_mul(Prec,Stat,F,G,ArgsL,ArgsR,Op,Rules0,RulesOut,Que):-
    %el 0 indica que si no = i hi ha grounds --> no separar sets
	sets_terms_disj_vars(ArgsL,ArgsR,SetsTLR,Op,0),
	tractar_sets_disj(Prec,Stat,SetsTLR,F,G,Op,Rules0,RulesOut,Que),!.


tractar_sets_disj(_,_,[[ArgsL,ArgsR]],F,G,Op,Rules0,[New_Rule|Rules0],ok):-!,
	L=..[F|ArgsL],
	R=..[G|ArgsR],
	New_Rule=..[Op|[L,R]],!.
tractar_sets_disj(_,_,Sets,F,G,'=',Rules0,RulesOut,ok):-!,
	tractar_sets_mul(F,G,Sets,'=',Rules0,RulesOut),!.
tractar_sets_disj(Prec,Stat,Sets,F,G,'>=',Rules0,RulesOut,ok):-!,
	sets_gr_segur_ge(Sets,Prec,Stat,Setsgr,Setsge,_),
	tractar_sets_mul(F,G,Setsgr,'>',Rules0,Rules1),
	tractar_sets_mul(F,G,Setsge,'>=',Rules1,RulesOut),!.
tractar_sets_disj(Prec,Stat,Sets,F,G,'>',Rules0,RulesOut,Que):-!,
	sets_gr_segur_ge(Sets,Prec,Stat,Setsgr,Setsge,GR),
	( 
 	 (GR=yes,Que=ok,
	  tractar_sets_mul(F,G,Setsgr,'>',Rules0,Rules1),
	  tractar_sets_mul(F,G,Setsge,'>=',Rules1,RulesOut)
	 );
	% si no sabem que algun segur > no podem generar subsets
	% fer tots > pot ser massa restrictiu. Fer tots >= en conjunt pot no ser >
	% escollir-ne un > foraria a fer backtracking.
	 (GR=no,Que=no, RulesOut=Rules0)
	),!.

	
tractar_sets_mul(_,_,[],_,Rules,Rules):-!.
tractar_sets_mul(F,G,[[ArgsL,ArgsR]|Sets],'=',Rules0,RulesOut):-!,
	length21(ArgsL,N),length21(ArgsR,M),!,N=M,
	L=..[F|ArgsL], R=..[G|ArgsR],
	New_Rule=..[_|[L,R]],
	tractar_sets_mul(F,G,Sets,'=',[New_Rule|Rules0],RulesOut).
tractar_sets_mul(F,G,[[ArgsL,ArgsR]|Sets],Op,Rules0,RulesOut):-
	L=..[F|ArgsL], R=..[G|ArgsR],
	New_Rule=..[Op|[L,R]],
	tractar_sets_mul(F,G,Sets,Op,[New_Rule|Rules0],RulesOut).


sets_gr_segur_ge([],_,_,[],[],GR):-!,(GR==yes;GR=no),!.
sets_gr_segur_ge([[AL,AR]|Sets],Prec,Stat,S1,S2,yes):-
	gr_multiset(AL,AR,Prec,Stat),!,
	sets_gr_segur_ge(Sets,Prec,Stat,S1,S2,yes),!.
sets_gr_segur_ge([[AL,AR]|Sets],Prec,Stat,[[AL,AR]|S1],S2,yes):-
	length21(AL,N),
	length21(AR,M),
	\+(M=N),!,
	sets_gr_segur_ge(Sets,Prec,Stat,S1,S2,yes),!.
sets_gr_segur_ge([[AL,AR]|Sets],Prec,Stat,S1,[[AL,AR]|S2],GR):-
	sets_gr_segur_ge(Sets,Prec,Stat,S1,S2,GR),!.

%%%%%%%%%%
%% S'usa quan lex o lexr
gen_ig_args([],[],_,_,Rules,Rules):-!.
gen_ig_args([A|_],[B|_],Prec,Stat,_,_):-
	segur_no_iguals(A,B,Prec,Stat),!,fail.
gen_ig_args([A|ArgsL],[B|ArgsR],Prec,Stat,Rules0,RulesOut):-
	gen_ig_args(ArgsL,ArgsR,Prec,Stat,[A=B|Rules0],RulesOut),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Per aquells termes de ArgsL (Arg) que segur que no sn iguals a 
%% cap de ArgsR, elimino tots els termes de ArgsR t.q. Arg s >rpo1
%% Si hi ha algun terme que ho compleix-->ho indiquem amb GR=yes.
simp_subt_gr(ArgsL,ArgsR,Prec,Stat,ArgsROut,yes):-
	delete33(Arg,ArgsL,ArgsL1),
	segur_no_igual_all(Arg,ArgsR,Prec,Stat),!,
	elim_gr_rpo1(Arg,ArgsR,Prec,Stat,ArgsR1),!,
	simp_subt_gr(ArgsL1,ArgsR1,Prec,Stat,ArgsROut,yes),!.
simp_subt_gr(_,ArgsR,_,_,ArgsR,GR):-!,(GR==yes;GR=no).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


check_rules_rpo(Rules,Prec,Stat,RulesOut):-
	check_rules_rpo(Rules,Prec,Stat,[],RulesOut),!.

check_rules_rpo([],_,_,Rules,Rules):-!.
check_rules_rpo([L=R|Rules],Prec,Stat,Rules0,RulesOut):-
	eq_rpo1(L,R,Prec,Stat),!,
	check_rules_rpo(Rules,Prec,Stat,Rules0,RulesOut),!.
check_rules_rpo([L=R|_],Prec,Stat,_,_):-
	segur_no_iguals(L,R,Prec,Stat),!,fail. %inclou L o R variable i L\=R.
check_rules_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut):-
	Rule=..['='|_],!,
	check_rules_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut),!.
% a partir d'aqu regles > o >=
check_rules_rpo([L>=R|Rules],Prec,Stat,Rules0,RulesOut):-
	is_variable(L),!, L==R,
	check_rules_rpo(Rules,Prec,Stat,Rules0,RulesOut),!.
check_rules_rpo([Rule|_],_,_,_,_):-
	Rule=..[_|[L,_]],	%L es variable i no s igual a R, o >
	is_variable(L),!,fail.
check_rules_rpo([L>=R|Rules],Prec,Stat,Rules0,RulesOut):-
	ge_rpo1(L,R,Prec,Stat),!,    
	check_rules_rpo(Rules,Prec,Stat,Rules0,RulesOut),!.
check_rules_rpo([L>R|Rules],Prec,Stat,Rules0,RulesOut):-
	gr_rpo1(L,R,Prec,Stat),!,
	check_rules_rpo(Rules,Prec,Stat,Rules0,RulesOut),!.
check_rules_rpo([L>=R|_],Prec,Stat,_,_):-
	gr_rpo1(R,L,Prec,Stat),!,fail.
check_rules_rpo([L>R|_],Prec,Stat,_,_):-
	ge_rpo1(R,L,Prec,Stat),!,fail.
check_rules_rpo([Rule|_],Prec,Stat,_,_):-
	Rule=..[Op|[L,R]],
	L=..[F|ArgsL],
	R=..[G|ArgsR],
	igual(F,G,Prec),
	quin_stat(F,Stat,StatF),
	\+(check_stat(StatF,ArgsL,ArgsR,Prec,Stat,Op)),!,fail.
check_rules_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut):-
	Rule=..[_|[L,R]],
	L=..[F|ArgsL], R=..[G|_],
	member212(G>F,Prec),!,
	subterm_possible(ArgsL,R,Prec,Stat),!,
	check_rules_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut),!.
check_rules_rpo([Rule|Rules],Prec,Stat,Rules0,RulesOut):-
	Rule=..[_|[L,R]],
	obtain_vars(L,Lvars),
	obtain_vars(R,Rvars),!,
	difer(Rvars,Lvars,[]),!,
	check_rules_rpo(Rules,Prec,Stat,[Rule|Rules0],RulesOut),!.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

step_rpo([],Prec,Stat,Prec,Stat):-!.
step_rpo(Rules,Prec,Stat,PrecOut,StatOut):-   %cal permetre back
	step_rpo1(Rules,Prec,Stat,Rules1,Prec1,Stat1),     
	step_rpo2(Rules1,Prec1,Stat1,Rules2,Prec2,Stat2),
	step_rpo3(Rules2,Prec2,Stat2,Rules3,Prec3,Stat3),
	step_rpo4(Rules3,Prec3,Stat3,PrecOut,StatOut).

%%
step_rpo12(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	step_rpo1(Rules,Prec,Stat,Rules1,Prec1,Stat1),
	step_rpo2(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut).

step_rpo123(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	step_rpo1(Rules,Prec,Stat,Rules1,Prec1,Stat1),  
	step_rpo2(Rules1,Prec1,Stat1,Rules2,Prec2,Stat2),
	step_rpo3(Rules2,Prec2,Stat2,RulesOut,PrecOut,StatOut).	


%%
%% per les regles amb tops F>=G : provem de fixar > i si fracassa =

step_rpo1([],Prec,Stat,[],Prec,Stat):-!.
step_rpo1(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	member22(Rule,Rules),
	Rule=..[_|[L,R]], L=..[F|_], R=..[G|_],
	member212(F>=G,Prec),!,
	(generarF(F>G,Prec,Prec1,Stat,Stat1);
	 generarF(F=G,Prec,Prec1,Stat,Stat1)),
	apply_simplifications(Rules,Prec1,Prec2,Stat1,Stat2,Rules2),
	step_rpo1(Rules2,Prec2,Stat2,RulesOut,PrecOut,StatOut).
step_rpo1(Rules,Prec,Stat,Rules,Prec,Stat).


%%
%% per les regles amb tops G>F : provem de fixar subterme (back)
step_rpo2([],Prec,Stat,[],Prec,Stat):-!.
step_rpo2(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules0),
	Rule=..[_|[L,R]], L=..[F|ArgsL], R=..[G|_],
	member212(G>F,Prec),!,
	obtenir_possible_subt(ArgsL,R,Sub,Prec,Stat),	%backtracking
	apply_simplifications([Sub>=R|Rules0],Prec,Prec1,Stat,Stat1,Rules1),
	call_step_rpo2(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut).
step_rpo2(Rules,Prec,Stat,Rules,Prec,Stat).

call_step_rpo2(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut):-
	member22(P,Prec1),P=..[>=|_],!,
	step_rpo12(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut).
call_step_rpo2(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut):-
	step_rpo2(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut).

%%
%% per les regles amb tops no relacionat provem de fixar F>G
%% per tops no relacionats i fracassa > provem subterme i si fracassa fixem F=G
%% per tops ='s provem subterme
%% Quan sortim d'aquest pas totes les regles tenen top =.
step_rpo3([],Prec,Stat,[],Prec,Stat):-!.
step_rpo3(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	member22(Rule,Rules),
	Rule=..[_|[L,R]], functor(L,F,_), functor(R,G,_),
	\+(rel(F,G,Prec)),
	generarF(F>G,Prec,Prec1,Stat,Stat1),
	apply_simplifications(Rules,Prec1,Prec2,Stat1,Stat2,Rules2),
	step_rpo123(Rules2,Prec2,Stat2,RulesOut,PrecOut,StatOut).
step_rpo3(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules0),
	Rule=..[_|[L,R]], L=..[F|ArgsL], functor(R,G,_),
	igual(F,G,Prec),
	obtenir_possible_subt(ArgsL,R,Sub,Prec,Stat),
	apply_simplifications([Sub>=R|Rules0],Prec,Prec1,Stat,Stat1,Rules1),
	step_rpo123(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut).
step_rpo3(Rules,Prec,Stat,RulesOut,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules0),
	Rule=..[_|[L,R]], functor(L,F,_), functor(R,G,_),
	\+(rel(F,G,Prec)),!,
	step3_subterm_or_fix_igprec(Rule,F,G,Prec,Stat,PrecOut,StatOut,Rules0,RulesOut).
step_rpo3(Rules,Prec,Stat,Rules,Prec,Stat).


step3_subterm_or_fix_igprec(Rule,F,_,Prec,Stat,PrecOut,StatOut,Rules,RulesOut):-
	Rule=..[_|[L,R]], L=..[F|ArgsL],
	obtenir_possible_subt(ArgsL,R,Sub,Prec,Stat),
	apply_simplifications([Sub>=R|Rules],Prec,Prec1,Stat,Stat1,Rules1),
	step_rpo123(Rules1,Prec1,Stat1,RulesOut,PrecOut,StatOut).
step3_subterm_or_fix_igprec(Rule,F,G,Prec,Stat,PrecOut,StatOut,Rules,RulesOut):-
	generarF(F=G,Prec,Prec1,Stat,Stat1),!,
	apply_simplifications([Rule|Rules],Prec1,Prec2,Stat1,Stat2,Rules2),
	step_rpo123(Rules2,Prec2,Stat2,RulesOut,PrecOut,StatOut).

	
%%
%% Totes les regles que queden tenen top = i sabem que no podem aplicar
%% cas subterme --> cal aplicar l'extensi segons Stat (si fixat).
%% Tamb sabem que amb el que hi ha fixat no rpo1.

step_rpo4([],Prec,Stat,Prec,Stat):-!.
% regla estricta amb lex aplicar def (opc. back)
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules1),
	Rule=..['>'|[L,R]],
	L=..[F|ArgsL], quin_stat(F,Stat,lex),!,
	R=..[_|ArgsR],  
 	simpl_args(lex,Prec,Stat,ArgsL,ArgsR,ArgsL1,ArgsR1),
	gen_grpo_lex(Prec,Stat,L,ArgsL1,ArgsR1,Rules1,Rules2),
	apply_simplifications(Rules2,Prec,Prec1,Stat,Stat1,Rules3),
	step_rpo(Rules3,Prec1,Stat1,PrecOut,StatOut).
% regla estricta amb lexr aplicar def (opc. back)
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules1),
	Rule=..['>'|[L,R]],
	L=..[F|ArgsL], quin_stat(F,Stat,lexr),!,
	R=..[_|ArgsR],  
 	simpl_args(lexr,Prec,Stat,ArgsL,ArgsR,ArgsL1,ArgsR1),
	reverse1(ArgsL1,ArgsL2),reverse1(ArgsR1,ArgsR2),
	gen_grpo_lex(Prec,Stat,L,ArgsL2,ArgsR2,Rules1,Rules2),
	apply_simplifications(Rules2,Prec,Prec1,Stat,Stat1,Rules3),
	step_rpo(Rules3,Prec1,Stat1,PrecOut,StatOut).
% regla >= amb lex o lexr aplicar > i si fracassa =
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules1),
	Rule=..['>='|[L,R]],
	functor(L,F,_), quin_stat(F,Stat,StatF),
	(StatF=lex;StatF=lexr),!,
	(New_Rule=..['>'|[L,R]];New_Rule=..['='|[L,R]]),
	apply_simplifications([New_Rule|Rules1],Prec,Prec1,Stat,Stat1,Rules2),
	step_rpo(Rules2,Prec1,Stat1,PrecOut,StatOut).
% = amb lex/lexr ja s'elimina a simplify_applic_rpo. Per tant,
% a partir d'aqu mul o cap
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules1),
	Rule=..['='|[L,R]],
	L=..[F|ArgsL], quin_stat(F,Stat,mul),!,
	R=..[G|ArgsR],
	delete33(TermL,ArgsL,ArgsL1),!,    %a partir d'aqu no back ja que han de ser =  
	delete33(TermR,ArgsR,ArgsR1),
	iguals_possible(TermL,TermR,Prec,Stat),
	L1=..[F|ArgsL1],R1=..[G|ArgsR1],
	apply_simplifications([TermL=TermR,L1=R1|Rules1],Prec,Prec1,Stat,Stat1,Rules2),
	step_rpo(Rules2,Prec1,Stat1,PrecOut,StatOut).
% regla > o >= amb mul intentem fer subtermes igual (opc. back)
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules1),
	Rule=..[Op|[L,R]],
	L=..[F|ArgsL], quin_stat(F,Stat,mul),
	R=..[G|ArgsR],
	delete33(TermL,ArgsL,ArgsL1),
	delete33(TermR,ArgsR,ArgsR1),
	iguals_possible(TermL,TermR,Prec,Stat),
	L1=..[F|ArgsL1],R1=..[G|ArgsR1],
	New_Rule=..[Op|[L1,R1]],
	apply_simplifications([TermL=TermR,New_Rule|Rules1],Prec,Prec1,Stat,Stat1,Rules2),
	step_rpo(Rules2,Prec1,Stat1,PrecOut,StatOut).
% regla > o >= amb mul segur cap subterme igual (opc. back) (per tant ser >)
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	delete33(Rule,Rules,Rules1),
	Rule=..[Op|[L,R]],
	L=..[F|ArgsL], quin_stat(F,Stat,mul),!,
	R=..[G|ArgsR],
% Sigui > o >=, segur que tots sn >.
% Puc eliminar les variables de ArgsL.
	elim_vars_ArgsL(ArgsL,ArgsL1),
	ArgsL1=[_|_],
	sets_terms_disj_vars(ArgsL1,ArgsR,Sets,Op,1),
	elim_sets_noR(Sets,Sets1),
	tractar_sets_rpo4(Prec,Stat,Sets1,L,R,F,G,Rules1,Rules2),
	apply_simplifications(Rules2,Prec,Prec1,Stat,Stat1,Rules3),
	step_rpo(Rules3,Prec1,Stat1,PrecOut,StatOut).
%regles amb stat=cap
step_rpo4(Rules,Prec,Stat,PrecOut,StatOut):-
	member22(Rule,Rules),
	Rule=..[Op|[L,R]],!,
	L=..[F|ArgsL], 	R=..[_|ArgsR],
%quin_stat(F,Stat,cap),
	fixar_stat(ArgsL,ArgsR,Prec,Stat,Stat1,F,Op),
	apply_simplifications(Rules,Prec,Prec1,Stat1,Stat2,Rules2),
	step_rpo(Rules2,Prec1,Stat2,PrecOut,StatOut).


%%%%%%%%%%%%%%%%
% un sol set i AL un sol element
tractar_sets_rpo4(_,_,[[[TermL],AR]],_,_,_,_,Rules,RulesOut):-!,
	gen_rules_rec_call(>,TermL,AR,Rules,RulesOut).
% un sol set
tractar_sets_rpo4(Prec,Stat,[[AL,AR]],L,_,_,G,Rules,RulesOut):-!,
	delete32(TermR,AR,AR1),
	obtenir_possible_subt(AL,TermR,TermL,Prec,Stat),
	((AR1=[], RulesOut=[TermL>TermR|Rules])
	 ;
	 (AR1=[_|_],  
	  afegir_rules_TermL(TermL,L,AR1,G,[TermL>TermR|Rules],RulesOut,Prec,Stat))
	).

% varis sets i subtermes ground a R (si ground sempre s el primer Set)
% L amb un sol argument --> Per tant sn 2 sets i els dos amb un sol argument
tractar_sets_rpo4(Prec,Stat,[[[TL],SetR],[[TL],SR2]],L,R,F,G,Rules,RulesOut):-
	delete32(T,SetR,SetR1),
	check_ground(T),!,
	concat312(SetR,SR2,SR),
	gen_rules_rec_call(>,TL,SR,Rules,RulesOut).
% varis sets i subtermes ground a R (si ground sempre s el primer Set)
% SetL amb varis elements
tractar_sets_rpo4(Prec,Stat,[[SetL,SetR]|Sets],L,R,F,G,Rules,RulesOut):-
	delete32(T,SetR,SetR1),
	check_ground(T),!,
	obtenir_possible_subt(SetL,T,TL,Prec,Stat),
	obtenir_setR_gr_possible_1(TL,SetR1,Sets,Argr,Prec,Stat),
	generar_subsets(Argr,OptArgr),  
	member22(Argr1,OptArgr),
	R=..[_|AR],
	difer(AR,[T|Argr1],AR1),
	gen_rules_rec_call(>,TL,[T|Argr1],Rules,Rules1),
	((AR1=[], RulesOut=Rules1)
	 ;
	 (AR1=[_|_], R1=..[G|AR1], 
	  L=..[_|ArgsL],
	  difer(ArgsL,TL,ArgsL1),
	  L1=..[F|ArgsL1],
	  RulesOut=[L1>R1|Rules1])
	).
% varis sets i NO subtermes a R ground
tractar_sets_rpo4(_,_,Sets,_,_,F,G,Rules,RulesOut):-
  	gen_gr_sets(Sets,F,G,Rules,RulesOut),!.

gen_gr_sets([],_,_,Rules,Rules):-!.
gen_gr_sets([[AL,AR]|Sets],F,G,Rules,RulesOut):-
	L=..[F|AL], R=..[G|AR],
	gen_gr_sets(Sets,F,G,[L>R|Rules],RulesOut).
	    % en aquest cas aquesta separaci no modifica el fet que amb iguals fracassa

%%%%%%%%

elim_sets_noR([],[]):-!.
elim_sets_noR([[_,[]]|Sets],SetsOut):-!,elim_sets_noR(Sets,SetsOut),!.
elim_sets_noR([[[],_]|_],_):-!,fail.
elim_sets_noR([Set|Sets],[Set|SetsOut]):-elim_sets_noR(Sets,SetsOut),!.


elim_vars_ArgsL([],[]):-!.
elim_vars_ArgsL([X|L],LOut):-
	is_variable(X),!,
	elim_vars_ArgsL(L,LOut).
elim_vars_ArgsL([X|L],[X|LOut]):-
	elim_vars_ArgsL(L,LOut).


% AR cont tots els arguments de R menys el TermR que ja hem tractat
% si a AR hi ha algun terme que pot ser igual a TermL,
% decidim quins termes de AR sn menors que TermL i treiem
% tots aquests termes de AR i TermL de ArgsL
afegir_rules_TermL(TermL,L,AR,G,Rules,RulesOut,Prec,Stat):-
	member22(TermR,AR),
	iguals_possible(TermL,TermR,Prec,Stat),!,
	obtenir_setR_gr_possible(TermL,AR,Argr,Prec,Stat),
	generar_subsets(Argr,OptArgr), 
	member22(Argr1,OptArgr), %Argr1 cont una llista de termes que 
	difer(AR,Argr1,AR1),	 %farem menors que TermL
	gen_rules_rec_call(>,TermL,Argr1,Rules,Rules1),
	((AR1=[], RulesOut=Rules1)
	 ;
	 (AR1=[_|_], R1=..[G|AR1], 
	  L=..[F|ArgsL],
	  difer(ArgsL,TermL,ArgsL1),
	  L1=..[F|ArgsL1],
	  RulesOut=[L1>R1|Rules1])
	).
% Si a AR cap igual possible a TermL llavors generem L>G(AR)
afegir_rules_TermL(_,L,AR,G,Rules,RulesOut,_,_):-
	R1=..[G|AR], 
	RulesOut=[L>R1|Rules].



% obt llista dels termes de AR que poden ser menors que T
obtenir_setR_gr_possible(T,AR,TgrAR,Prec,Stat):-
	findall(TR, obtenir_gr_possible(T,AR,Prec,Stat,TR),TgrAR).

% considera gr_possible tenint en compte variables i
% que no sigui Term>=T
obtenir_gr_possible(T,LTerms,Prec,Stat,Term):-
	obtain_vars(T,Tvars),
	member22(Term,LTerms),
	obtain_vars(Term,Rvars),
	difer(Rvars,Tvars,[]),
	\+(ge_rpo1(T,Term,Prec,Stat)).

%si T s ground noms pot ser ms gran que altres ground
obtenir_setR_gr_possible_1(T,ARground,_,TgrAR,Prec,Stat):-
	check_ground(T),!,
	obtenir_setR_gr_possible(T,ARground,TgrAR,Prec,Stat).
%si T NO s ground cal considerar altres ground i els del set on apareix T
obtenir_setR_gr_possible_1(T,ARground,Sets,TgrAR,Prec,Stat):-
	buscar_set_T(T,Sets,ART),
	concat312(ARground,ART,ARgrT),
	obtenir_setR_gr_possible(T,ARgrT,TgrAR,Prec,Stat).



buscar_set_T(_,[],[]):-!.
buscar_set_T(T,[[AL,ART]|_],ART):-
	member21(T,AL),!.
buscar_set_T(T,[_|Sets],ART):-
	buscar_set_T(T,Sets,ART).

check_ground([]):-!.
check_ground([S|R]):-!, check_ground(S),!,
		     check_ground(R).
check_ground(S):- is_variable(S),!,fail.
check_ground(S):- S=..[_|Args],
		  check_ground(Args).


generar_subsets(Set,LSubsets):-
	delete32(T,Set,Set1),
	generar_subsets(Set1,LSub),
	map(cons,T,LSub,LSub1),
	concat312(LSub,LSub1,LSubsets).
generar_subsets([],[[]]).


%%%%%%%%%%%%%%%%	

gen_grpo_lex(_,_,_,[_|_],[],Rules,Rules):-!.
gen_grpo_lex(_,_,L,[A|_],[B|ArgsR],Rules,RulesOut):-
	gen_rules_rec_call(>,L,ArgsR,[A>B|Rules],RulesOut).
gen_grpo_lex(Prec,Stat,L,[A|ArgsL],[B|ArgsR],Rules,RulesOut):-
	gen_grpo_lex(Prec,Stat,L,ArgsL,ArgsR,[A=B|Rules],RulesOut).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%s'usa des de step_rpo4(1) i des de tractar_args_mul(0).
sets_terms_disj_vars(TermsL,TermsR,SetsTLR,Op,Fixed):-
	concat312(TermsL,TermsR,T),
	set_disj_vars(T,[],SetsVars),
	set_terms_disj_vars(TermsL,SetsVars,[],SetsTermsL,TgroundL),	
        set_terms_disj_vars(TermsR,SetsVars,[],SetsTermsR,TgroundR),
	zip(list2,SetsTermsL,SetsTermsR,SetsLR),
	concl(Op,TgroundL,TgroundR,SetsLR,TermsL,TermsR,SetsTLR,Fixed),!.


concl('=',TgroundL,TgroundR,SetsLR,_,_,SetsOut,_):-!,SetsOut=[[TgroundL,TgroundR]|SetsLR].
concl(_,_,[],SetsLR,_,_,SetsOut,_):-!,SetsOut=SetsLR.
concl(_,_,TgroundR,SetsLR,TermsL,_,SetsOut,1):-!,SetsOut=[[TermsL,TgroundR]|SetsLR].
concl(_,_,_,_,TermsL,TermsR,[[TermsL,TermsR]],0):-!.


set_terms_disj_vars(Terms,[],SetsTerms,SetsTerms,Terms):-!.
set_terms_disj_vars(Terms,[SetVars|SetsVars],SetsTerms,SetsTermsOut,Tground):-
	obtain_terms_set(Terms,SetVars,TermsSet,RestTerms),
	set_terms_disj_vars(RestTerms,SetsVars,[TermsSet|SetsTerms],SetsTermsOut,Tground),!.

obtain_terms_set([],_,[],[]):-!.
obtain_terms_set([T|R],Vars,[T|R1],R2):-
	obtain_vars(T,VT),
	member22(V,VT),member22(V,Vars),!,
	obtain_terms_set(R,Vars,R1,R2),!.
obtain_terms_set([T|R],Vars,R1,[T|R2]):-
	obtain_terms_set(R,Vars,R1,R2),!.

set_disj_vars([T|R],Sets,SetsOut):-
	obtain_vars(T,VarsT), VarsT=[_|_],!,
	afegir_disj_vars(VarsT,R,VarsSet,[],R1),
	set_disj_vars(R1,[VarsSet|Sets],SetsOut),!.
set_disj_vars([_|R],Sets,SetsOut):-
	set_disj_vars(R,Sets,SetsOut).
set_disj_vars([],Sets,Sets):-!.


afegir_disj_vars(Vars,[X|R],VarsSet,R1,ROut):-
	obtain_vars(X,VX),
	member22(V,VX),
	member22(V,Vars),!,
	concat312(Vars,VX,Vars1),
	sort(Vars1,Vars11),
	concat312(R,R1,RR),
	afegir_disj_vars(Vars11,RR,VarsSet,[],ROut),!.
afegir_disj_vars(Vars,[X|R],VarsSet,R1,ROut):-
	afegir_disj_vars(Vars,R,VarsSet,[X|R1],ROut),!.
afegir_disj_vars(Vars,[],Vars,R,R):-!.



