%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% order.pl 	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-dynamic interactive/0.
:-dynamic innermost/0.
:-dynamic deeper_analysis/0.

:-dynamic(ajuda/1). 
:-dynamic(ajuda2/1). 
:-dynamic(nosat/1). 
:-dynamic(nosetsat/1). 
:-dynamic(nosatinn/2). 

:-dynamic(ok_rulesSat/0). 
:-dynamic(no_rulesSat/0). 


ccStand:-deeper_analysis,!.
ccStand:-assertz(deeper_analysis).
noccStand:-retractall(deeper_analysis).
inter:-interactive,!.
inter:-assertz(interactive).
nointer:-retractall(interactive).
inner:-innermost,!.
inner:-assertz(innermost).
noinner:-retractall(innermost).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

inici(Rules):-
	clean,
	check_rules_and_signature(Rules,Def,Const),
	obtain_vars_system(Rules,Vars0),
        check_no_name_conflict(Def,Const,Vars0,Vars),
	assertz(leftR(Def)),
	assertz(noleftR(Const)),
	assertz(variables(Vars)),
	assertz(leftR0(Def)),
	assertz(noleftR0(Const)),
	check_rules_wf(Rules),!.

clean:-	retractall(leftR(_)),
	retractall(noleftR(_)),
	retractall(variables(_)),
	retractall(leftR0(_)),
	retractall(noleftR0(_)),
	retractall(renamed(_)),
	retractall(ajuda(_)),
	retractall(ajuda2(_)),
	retractall(nosat(_)),
	retractall(nosetsat(_)),
	retractall(nosatinn(_,_)),
	retractall(ok_rulesSat),
	retractall(no_rulesSat),
	retractall(maxAssert(_)),
	retractall(q(_)),!.

clean_end:- clean,
	    clean_status,!.

clean_status:-
	retractall(interactive),
	retractall(statuslex),
	retractall(innermost),
	retractall(deeper_analysis),!.


limit_asserts(Rules):-
	leftR0(Def),
	noleftR0(Cons),
	concat312(Def,Cons,Symbs),
	limit_asserts(Rules,Symbs),!.

limit_asserts(Rules,Symbs):-
	length21(Rules,NumRules),	
	obt_max_arity(Symbs,0,MaxAr),
	obt_max_allowed(NumRules,MaxAr,NumMax),
	retractall(maxAssert(_)),
	assertz(maxAssert(NumMax)),
	retractall(q(_)),
	assertz(q(0)),!.

obt_max_arity([],A,A):-!.
obt_max_arity([[_,X]|Symbs],A,MaxAr):-
	X>A,!,
	obt_max_arity(Symbs,X,MaxAr),!.
obt_max_arity([_|Symbs],A,MaxAr):-
	obt_max_arity(Symbs,A,MaxAr),!.


obt_max_allowed(NumRules,MaxAr,30):-
	NumRules=<10, 
	MaxAr=<2,!.
obt_max_allowed(NumRules,MaxAr,20):-
	(NumRules=<10 
	 ; 
	 (NumRules=<20,	MaxAr=<2)
	),!.
obt_max_allowed(NumRules,_,15):-
	NumRules=<20,!.
obt_max_allowed(_,_,10).


check_limit_Asserts:-
	retract(q(K)), 
	maxAssert(M), 
	((K<M,!, K1 is K+1)
	 ;
	 (K1=1, 
	  retractall(nosat(_)),
	  retractall(nosetsat(_)),
	  retractall(nosatinn(_,_))
	 )
	),!,
	assertz(q(K1)),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ordre([],['yes'],_,_,[]):-!.
ordre(Rules,Resultat,H,Kind,[Rules1Enum,CamiIni,ConstIniSimpl1,LproofSubgraphs]):-
	inici(Rules),
	simplif_rules_strat(Rules,Rules1),
	limit_asserts(Rules1),
	enumerate(Rules1,Rules1Enum),
	obt_const_trs(Kind,Rules1Enum,LCAll),
	enumerate(LCAll,LCAllEnum),
	separ_cons(LCAllEnum,LCAllConstEnum),
  	triv_or_cont(LCAllConstEnum,Rules1,Resultat,H,[N,LproofSubgraphs]),!,
	member21([N,[ConstIniSimpl1,CamiIni]],LCAllEnum),!.
ordre(_,['no'],_,_,[]):-!.

%%%%%%%%
%kind is plain, twosteps, unaun or two_unaun.

obt_const_trs(plain,LRules,LCAll):-!,
	const_trs(LRules,LCAll).
obt_const_trs(twosteps,LRules,LCAll):-!,
	const_trs_2steps(LRules,LCAll).
obt_const_trs(unaun,LRules,[C]):-!,
	const_one_trs(LRules,C).
obt_const_trs(two_unaun,LRules,LC):-!,
	const_trs_hard1(LRules,LC).


%%%%%%%%%%%%%%%

separ_cons([],[]):-!.
separ_cons([[N,[Cons,_]]|Rest],[[N,Cons]|R1]):-
	separ_cons(Rest,R1),!.


%%%%%%%%%%%%%%%

obtain_3_first_constraints(LC,LC0):-
	LC=[_,_,_,_|_],!,  % ms de 3 constraints
	eval_each_constraint(LC,LC_Eval),
	sort(LC_Eval,LCOrdVal),!,
	separar_1_2(LCOrdVal,LCOrdVal1,LEnum),
	map(snd2,LCOrdVal1,LCOrd),
	zip(list2,LEnum,LCOrd,LCEnumOrd),
        sel_3(LCEnumOrd,LC0).
obtain_3_first_constraints(L,L).

eval_each_constraint(LC,LC_Eval):-
	separar_1_2(LC,LEnum,LC1),
	map(eval_cons,LC1,LC_Eval0),
	zip(list2,LC_Eval0,LEnum,LC_Eval),!.

eval_cons(Q,[Val,Q]):-
	eval_cons1(Q,0,0,Ngr,Nge),
	Val is 2*Ngr+Nge,!.
	
eval_cons1([],N1,N2,N1,N2):-!.
eval_cons1([_>_|Rest],N1,N2,M1,M2):-
	N11 is N1+1,	
	eval_cons1(Rest,N11,N2,M1,M2),!.
eval_cons1([_>=_|Rest],N1,N2,M1,M2):-
	N22 is N2+1,	
	eval_cons1(Rest,N1,N22,M1,M2),!.



sel_3(LC,LC):-length21(LC,N),N<4,!.
sel_3([X,Y,Z|_],[X,Y,Z]).
sel_3([_,_,_|LC],LC1):-
	sel_3(LC,LC1).


all_is_ge([]):-!.
all_is_ge(L):-
	allisge(L).

allisge([]):-!.
allisge([_>=_|R]):-allisge(R),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% si LC=[] vol dir que treient els imposibles [] --> no.
%%
triv_or_cont([],_,['no'],_,[]):-!.
triv_or_cont(LC,_,['triv1'],_,[N,[[],[],[]]]):-member21([N,[]],LC),!.
triv_or_cont(LC,_,['triv2',C],_,[N,[[],[],[]]]):- 
	member22([N,C],LC),
	all_is_ge(C),!.
triv_or_cont(LC,Rules,Resultat,H,[Num,[Def,V,LproofSubgraphs]]):-
	renvars_list1(Rules,Rules1),
	convert_rules_const(Rules1,C_Rules),
	obtain_3_first_constraints(LC,LC0),
	conv_cts_vars(LC0,LC1),
 	list_const_cycles(Rules,LC1,List_C_Cycles),
  	(( try_cancel_cycles,
	   constriv_or_cont_inn(List_C_Cycles,C_Rules,Resultat,H,Num,LProof),
	   member21([Num,L],List_C_Cycles),
	   obt_Lproofsubgraphs(L,LProof,LproofSubgraphs),
	   leftR0(Def), variables(V)	 	
	 )
	 ;
	 ( \+ try_cancel_cycles,
	   constriv_or_cont(List_C_Cycles,C_Rules,Resultat,H,Num),
	   member21([Num,LproofSubgraphs],List_C_Cycles),  % LSubgrafs=[[Qenum,LStrict,LNoStrict],...]
	   Def=[],V=[]  
	 )				% Resultat mateix ordre que LSubgrafs - 1 per cada subgraf
	),!.
		
	  
try_cancel_cycles:-innermost.
try_cancel_cycles:-deeper_analysis.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% comprova si hi ha constraint que no contingui cap cicle amb almenys un >
constriv_or_cont(LQC,_,['triv3'],_,Num):-
	member22([Num,[]],LQC),!.
constriv_or_cont(LQC,C_Rules,['si1',Resultat],Help0,Num):-
	rulesSat(C_Rules),
% LQC=[[N,[[Qenum1,Ldcyc1,LnoS1],[Qenum2,Ldcyc2,LnoS2],...]],...] ; 
%     Qenum=[q(1,C1),q(2,C2),...] ;  Ldcyc=[[1],[1,3,4],[2,4]])
	map(elim_NoStricts,LQC,LQC1),
	map(obtain_dif_alt_gr_ge,LQC1,LAltSubgrafs),
% LAltSubgrafs=[[N,[[Qenum1,Laltdcyc1],[Qenum2,Laltdcyc2],...]],...] ; 
% Laltdcyc=[[1,2],[1,4]] 
% opcions de estrictes que asseguren  almenys un estricte en cada cicle del subgraf.)
	sel_firstQ_to_prove(LAltSubgrafs,[Num,QAltern]),
% QAltern=[[Qenum1,Laltdcyc1],[Qenum2,Laltdcyc2],...]
     	check_no_Set_setnosat(QAltern),
	ask_help0(Help0), 
	proof_for_every_disj_set(C_Rules,QAltern,Resultat,Help0),!.


%%%%%%%%%%
% rulesSat - comprova que les regles soles siguin satisfactibles 
%	     d'aqu endavant segur que hi ha algun constraint estricte 
%	     i per tant no val soluci trivial. UNA UNICA VEGADA.
%	(noms cal comprovar-ho si hi ha regles col.lapsants, sin s trivial
%	 que t soluci)
%	No cal dir quines regles sn. Abans de canviar el sistema de regles
%	es fa un clean.


rulesSat(_):-
	ok_rulesSat,!.
rulesSat(_):-
	no_rulesSat,!,fail.
rulesSat(CRules):-
	\+(exists_collapsing(CRules)),!,
	assertz(ok_rulesSat),!.
rulesSat(CRules):-
	prove_set_int_rpo(CRules,[],_,'n'),!,
	assertz(ok_rulesSat),!.
rulesSat(_):-
	assertz(no_rulesSat),!,fail.


exists_collapsing([_>=R|_]):-is_variable(R),!.
exists_collapsing([_|Rest]):-exists_collapsing(Rest),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

elim_NoStricts([N,LSub],[N,LSubS]):-
	map(elim_NoStricts1,LSub,LSubS).

elim_NoStricts1([Q,S,_],[Q,S]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sep_disj_cycles(LCC,LQDC):-
	simplify_Q_cycles(LCC,LCC1),
	separ_disjoint_cycles(LCC1,LQDC),!.

%% eliminem els constraints de Q que no intervenen en cap cicle

simplify_Q_cycles([],[]):-!.
simplify_Q_cycles([[N,Q,Lcyc]|Rest],[[N,Q1,Lcyc]|Rest1]):-
	l_union(Lcyc,Lnum),
	selec_Q(Q,Lnum,Q1),!,
	simplify_Q_cycles(Rest,Rest1),!.

selec_Q([],_,[]):-!.
selec_Q([q(N,C)|R],Lnum,[q(N,C)|R1]):-
	member21(N,Lnum),!,
	selec_Q(R,Lnum,R1),!.
selec_Q([_|R],Lnum,R1):-
	selec_Q(R,Lnum,R1),!.
	
%%%%
% separ_disjoint_cycles(LCCycles,LQdisjCycles) :

% PRECOND: cal que a tot q(N,CN) de Qenum surti en algun $cyc_i$  

% LCCycles=[Qenum,[cyc1,cyc2,...]],...] 
% 	Qenum=[q(1,C1),q(2,C2),...] ; (e.g. cyc1=[1,3,4])
% LQdisjCycles=[Qenum,[Ldcyc1,Ldcyc2,...]],...] ; 
%	Ldcyc1=[AllpairsDisjCycles,LcyclesSet] ;
%	e.g. Ldcyc1=[[1,2,3,4],[[1],[1,3,4],[2,4]])

separ_disjoint_cycles([],[]):-!.
separ_disjoint_cycles([LCC|LCCycles],[QDC|LQdisjCycles]):-
	separ_disjoint_cycles1(LCC,QDC),!,
	separ_disjoint_cycles(LCCycles,LQdisjCycles),!.

separ_disjoint_cycles1([N,Q,Lcyc],[N,Q,LDcycs]):-
	map(selnumQ,Q,Lnum), 
	separ_disjoint_cycles2(Lnum,Lcyc,LDcycs),!.
separ_disjoint_cycles1([],[]). %mai s'usa

selnumQ(q(N,_),N).
selconsQ(q(_,C),C).
	
separ_disjoint_cycles2([],_,[]):-!.
separ_disjoint_cycles2([N|Lnum],Lcyc,[[Ncyc,CycRelN]|Dcycs]):-
	innermost,!,
	obtain_rels_cycs_inn([N],Lnum,Lcyc,[],CycRelN,Ncyc,Lnum1,Lcyc1),!,
	separ_disjoint_cycles2(Lnum1,Lcyc1,Dcycs),!.
separ_disjoint_cycles2([N|Lnum],Lcyc,[[Ncyc,CycRelN]|Dcycs]):-
	obtain_rels_cycs([N],Lnum,Lcyc,[],CycRelN,Ncyc,Lnum1,Lcyc1),!,
	separ_disjoint_cycles2(Lnum1,Lcyc1,Dcycs),!.

obtain_rels_cycs(Ncyc,Lnum,Lcyc,CycRelN0,CycRelNOut,NcycOut,LnumOut,LcycOut):-
	delete33(Cyc,Lcyc,Lcyc1),
	inter(Cyc,Ncyc,[_|_]),!, % No disjunts
	difer(Lnum,Cyc,Lnum1),
	unio1(Ncyc,Cyc,Ncyc1),!,
	obtain_rels_cycs(Ncyc1,Lnum1,Lcyc1,[Cyc|CycRelN0],CycRelNOut,NcycOut,LnumOut,LcycOut),!.
obtain_rels_cycs(Ncyc,Lnum,Lcyc,Cyc,Cyc,NcycOut,Lnum,Lcyc):-sort(Ncyc,NcycOut).

obtain_rels_cycs_inn(Ncyc,Lnum,Lcyc,CycRelN0,CycRelNOut,NcycOut,LnumOut,LcycOut):-
	delete33([Lcancel|Cyc],Lcyc,Lcyc1),
	inter(Cyc,Ncyc,[_|_]),!, % No disjunts
	difer(Lnum,Cyc,Lnum1),
	unio1(Ncyc,Cyc,Ncyc1),!,
	obtain_rels_cycs_inn(Ncyc1,Lnum1,Lcyc1,[[Lcancel|Cyc]|CycRelN0],CycRelNOut,NcycOut,LnumOut,LcycOut),!.
obtain_rels_cycs_inn(Ncyc,Lnum,Lcyc,Cyc,Cyc,NcycOut,Lnum,Lcyc):-sort(Ncyc,NcycOut).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_no_Set_setnosat([]):-!.
check_no_Set_setnosat([[Qenum,_]|_]):-
	map(selconsQ,Qenum,Qreal), 
	sort(Qreal,QrealOrd),
	nosetsat(QrealOrd),!,fail.
check_no_Set_setnosat([_|R]):-
	check_no_Set_setnosat(R).


proof_for_every_disj_set(_,[],[],_):-!.	
proof_for_every_disj_set(C_Rules,[[Qenum,L_lgr]|LSets],[RtatSet|Resultat],Help):-	
	sort_len(L_lgr,L_lgrOrd),!,	
  % d'entre els que tenen el mateix Qenum fem primer els que tenen menys estrictes
	map(selconsQ,Qenum,Qreal),  % Qreal son els pairs originals sense enumerar 
	sort(Qreal,QrealOrd),
	prove_one_set(C_Rules,Qenum,QrealOrd,L_lgrOrd,RtatSet,Help),!,
	proof_for_every_disj_set(C_Rules,LSets,Resultat,Help),!.
	

%prove_one_set(_,_,Qreal,_,RtatSet,_):-
%	setsat(Qreal,RtatSet),!.
prove_one_set(C_Rules,Qenum,_,L_lgr,[LConsOrd|RtatSet],Help):-
	member22(Lgr,L_lgr),
	obtain_cycles1(Qenum,Lgr,LCons),
	sort(LCons,LConsOrd),
	prove_set_int_rpo(C_Rules,LConsOrd,RtatSet,Help),!.
%	prove_set_int_rpo(C_Rules,LConsOrd,RtatSet,Help),!,
%	assertz(setsat(Qreal,[LConsOrd|RtatSet])),!.
prove_one_set(_,_,Qreal,_,_,_):-
	ajuda2('n'),
	check_limit_Asserts,
	assertz(setnosat(Qreal)),!,fail.

%%%%%

sel_Lnum_Qenum([],_,[]):-!.
sel_Lnum_Qenum([q(N,C)|R],Lnum,[q(N,C)|R1]):-
	delete32(N,Lnum,Lnum1),!,
	sel_Lnum_Qenum(R,Lnum1,R1),!.
sel_Lnum_Qenum([_|R],Lnum,R1):-
	sel_Lnum_Qenum(R,Lnum,R1),!.

	
obtain_cycles1([],_,[]):-!.
obtain_cycles1([q(N,C)|Q],Lgr,[L>R|L_Q]):-
	delete32(N,Lgr,Lgr1),!,C=..[_|[L,R]],
	obtain_cycles1(Q,Lgr1,L_Q),!.
obtain_cycles1([q(_,C)|Q],Lgr,[L>=R|L_Q]):-
	C=..[_|[L,R]],
	obtain_cycles1(Q,Lgr,L_Q),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%prove_set_int_rpo(_,C,ResC,_):-
%	C=[_|_], %per no usar-ho quan regles soles
%	sat(C,ResC),!.  
prove_set_int_rpo(_,C,_,_):-
	C=[_|_], %per no usar-ho quan regles soles
	nosat(C),!,fail.  
prove_set_int_rpo(C_Rules,C_Cons,[IntfixOut,PrecOut,StatOut],Help):-
	prepare_int_rpo(C_Rules,C_Cons,Prec,Stat,Intfix,Help),
	try_set_int_rpo(C_Rules,C_Cons,Intfix,IntfixOut,Prec,PrecOut,Stat,StatOut).


try_set_int_rpo(C_Rules,C_Cons,Intfix,IntfixOut,Prec,PrecOut,Stat,StatOut):-
	int_rpo(C_Rules,C_Cons,Intfix,IntfixOut,Prec,PrecOut,Stat,StatOut),!.
%	int_rpo(C_Rules,C_Cons,Intfix,IntfixOut,Prec,PrecOut,Stat,StatOut),!,
%	assertz(sat(C_Cons,[IntfixOut,PrecOut,StatOut])).
try_set_int_rpo(_,C_Cons,_,_,_,_,_,_):-
	ajuda2('n'),
	check_limit_Asserts,
	assertz(nosat(C_Cons)),!,fail.
	


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%	 map(obtain_dif_alt_gr_ge,LQC,LAltSubgrafs)
%
%	obtain_dif_alt_gr_ge([N,LSubgrafs],
%
% LQC=[[N,LSubgrafs],...] ; LSubgrafs=[[Qenum1,Ldcyc1],[Qenum2,Ldcyc2],...]
%     Qenum=[q(1,C1),q(2,C2),...] ;  Ldcyc=[[1],[1,3,4],[2,4]]
% LAltSubgrafs=[[N,[[Qenum1,Laltdcyc1],[Qenum2,Laltdcyc2],...]],...] ; 
% Laltdcyc=[[1,2],[1,4]] =alternatives de conjunts de parells estrictes 
% que asseguren almenys un estricte en cada cicle. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


obtain_dif_alt_gr_ge([N,LSubgrafs],[N,LAlternSubgrafs]):-
	map(obtain_dif_alt_gr_ge0,LSubgrafs,LAlternSubgrafs),!.

obtain_dif_alt_gr_ge0([Qenum,Lcycsnum],[Qenum,LAltcycsnum]):-
	obtain_dif_alt_gr_ge1(Lcycsnum,LAltcycsnum),!.

obtain_dif_alt_gr_ge1(Lcycles,L_Lgr):-
	obtain_gr(Lcycles,Lcycles1,[],Lgr),!,
	obtain_rest_gr(Lcycles1,Lgr,L_Lgr),!.
		

obtain_gr(Lcycles,LcyclesOut,Lgr,LgrOut):-
	delete32([N],Lcycles,Lcycles1),
	elim_has_N(N,Lcycles1,Lcycles2),
	obtain_gr(Lcycles2,LcyclesOut,[N|Lgr],LgrOut),!.
obtain_gr(Lcycles,Lcycles,Lgr,Lgr):-!.
	

elim_has_N(_,[],[]):-!.
elim_has_N(N,[L|Lcycles],LcyclesOut):-
	member21(N,L),!,
	elim_has_N(N,Lcycles,LcyclesOut),!.
elim_has_N(N,[L|Lcycles],[L|LcyclesOut]):-
	elim_has_N(N,Lcycles,LcyclesOut),!.


obtain_rest_gr(Lcycles,Lgr,L_Lgr_Out):-
	l_union(Lcycles,Lnums),
	findall(L,sel_gr_ge(Lcycles,Lnums,L),L_Lgr),
	elim_Lgr_subsum(L_Lgr,L_Lgr1),
	map(concat312,Lgr,L_Lgr1,L_Lgr_Out),!.


sel_gr_ge([],_,[]).
sel_gr_ge(Lcycles,[N|Lnums],[N|L]):-
	elim_has_N(N,Lcycles,Lcycles1),
	l_union(Lcycles1,Lnumscycles1),
	inter(Lnums,Lnumscycles1,Lnums1),
	sel_gr_ge(Lcycles1,Lnums1,L).
sel_gr_ge(Lcycles,[_|Lnums],L):-
	sel_gr_ge(Lcycles,Lnums,L).

%%%	
elim_Lgr_subsum(L,LOut):-
	elim_Lgr_subsum1(L,L,LOut),!.

elim_Lgr_subsum1([],L,L):-!.
elim_Lgr_subsum1([X|Rest],L,LOut):-
	elim_Lgr_subsum11(X,L,L1,Elim),
	difer(Rest,Elim,Rest1),	
	elim_Lgr_subsum1(Rest1,[X|L1],LOut).

elim_Lgr_subsum11(_,[],[],[]):-!.
elim_Lgr_subsum11(X,[Y|L],L1,[Y|Elim]):-
	difer(X,Y,[]),!,		%Y subsumit a X
	elim_Lgr_subsum11(X,L,L1,Elim),!.
elim_Lgr_subsum11(X,[Y|L],[Y|L1],Elim):-
	elim_Lgr_subsum11(X,L,L1,Elim),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   sel_firstQ_to_prove(LConstraints,[N,LAltsSubgrafs])

sel_firstQ_to_prove([Qsel],Qsel):-!. % noms n'hi ha un.
sel_firstQ_to_prove(LConstraints,Qsel):-
	map(obtain_val_heur,LConstraints,LConstraints1),!,
	sort(LConstraints1,LAltminFirst),!, %ordena de menor a major per la primera
					  %component de l'estructura heur(H,SH,_,_).
	member22(heur(_,_,N,LAltsSubgrafs),LAltminFirst),  %backtracking
	Qsel=[N,LAltsSubgrafs].

%%%
% obt dos valors:
%	a. el valor maxim d'entre tots els conjunts de cicles disjunts 
%	b. la suma de valors de cada conjunt de cicles disjunt
% El valor d'un conjunt de cicles s'obt de la millor alternativa (minim valor)
%	sumant els pairs que intervenen ponderant: >= val 1 i > val 2.

obtain_val_heur([N,LAltsSubgrafs],heur(Heur,SHeur,N,LAltsSubgrafs)):-
	obtain_val_heur1(LAltsSubgrafs,LVals),!,
	max_list(LVals,Heur),
	sum(LVals,SHeur),!.
	
obtain_val_heur1([[Qenum,L_lgr]|R],[Val|R1]):-
	length21(Qenum,N1),
	min_length(L_lgr,N2),
	Val is N1+N2,!,
	obtain_val_heur1(R,R1),!.
obtain_val_heur1([],[]):-!.

min_length([Lgr|L_lgr],MinLOut):-
	length21(Lgr,MinL),
	min_length1(MinL,L_lgr,MinLOut),!.

min_length1(MinL,[],MinL):-!.
min_length1(MinL,[L|Rest],MinLOut):-
	length21(L,N),
	MinL>N,!,
	min_length1(N,Rest,MinLOut),!.
min_length1(MinL,[_|Rest],MinLOut):-
	min_length1(MinL,Rest,MinLOut),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%  Per INNERMOST i DEEPER_ANALYSIS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

constriv_or_cont_inn(LQC,_,['triv3'],_,N,[]):-
	member22([N,[]],LQC),!.
constriv_or_cont_inn(LQC,C_Rules,['si2',Resultat],Help0,Num,Lproof):-
	rulesSat(C_Rules),
 % ATENCI : si en algun moment podem fer una millora com que noms
 % 	     cal considerar les regles que poden intervenir en els cicles
 % 	     cal eliminar la comprovaci rulesSat (noms cas Innermost ja
 % 	     que les variables estan en NF)
% LQC=[[N,[Subgraf1,Subgraf2,...]],...] 
%     Subgraf1=[Qenum,Lcycstrict,LcycNostrict,[P1,P2,...]]  
%    Pi=['a1',Nums,Ldcyc] or ['a2',Nums,Ldcyc]  with
%     Qenum=[q(1,C1),q(2,C2),...] ; Ldcyc=[[Lc1,1],[Lc2,1,3,4],[Lc3,2,4]]
% or Pi=['b','i',Lconds,LcancelledCycs,LSubgrafs] with 
%	LSubgrafs=[S1,S2,...] Si=[K,Nums,Lstrict,LNostrict] (K='a1' or 'a2')
 	map(obtain_dif_alt_inn,LQC,LAltCycles),
% LAltCycles=[ (1)[N1,[Alts_Sub1, Alts_Sub2,...]], (2)...] ; 
% Alts_Sub1=[Qenum,[Alt1,Alt2,...]] ; 
% Alt1=['a1' or 'a2,[Laltdcyc1,Laltdcyc2,...]] with Laltdcyc1=[Lc1,Lgr1,Lge1] 
% or  =['b','i',LAltsSubgraphs] ; LAltsSubgraphs=[Alt11,Alt12,...] (tipus 'a1' i 'a2')
	sel_firstQ_inn_to_prove(LAltCycles,[Num,QSel]),
% [Num,QSel]=[Num , [[Qenum1,Alts_Sub1], [Qenum2,Alts_Sub2],...] ]
%
% De moment ho eliminem:  check_inn_no_Set_setnosat(QSel),
	ask_help0(Help0),
	proof_for_every_subgraph_inn(C_Rules,QSel,Lproof,Resultat,Help0),!.

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%sep_disj_cycles_inn(LQC,LQdisjCycles):-
%	simplify_Qcycles_inn(LQC,LQC1),
%	separ_disjoint_cycles(LQC1,LQdisjCycles),!.

%simplify_Qcycles_inn([],[]):-!.
%simplify_Qcycles_inn([[N,Q,Lcyc]|LQC],[[N,Q1,Lcyc]|LQCOut]):-
%	map(tail,Lcyc,L_Lnums),
%	l_union(L_Lnums,Nums),
%	selec_Q(Q,Nums,Q1),!,
%	simplify_Qcycles_inn(LQC,LQCOut),!.
	
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%separ_Qenum([N,Qenum,Lcycs],[N,LOut]):-
%	map(selQLnum,Qenum,Lcycs,LOut),!.

%selQLnum(Qenum,[Lnum,Lcycs],[Qenum1,Lcycs]):-
%	selec_Q(Qenum,Lnum,Qenum1),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% LAltCycles=[ (1)[N1,[Alts_Sub1, Alts_Sub2,...]], (2)...] ; 
% Alts_Sub1=[Qenum,[Alt1,Alt2,...]] ; 
% Alt1=['a1' or 'a2',[Laltdcyc1,Laltdcyc2,...]] with Laltdcyc1=[Lc1,Lgr1,Lge1] 
% or  =['b',I,LAltsSubgraphs] ; LAltsSubgraphs=[Alt11,Alt12,...] (noms tipus 'a?')

% [N,Qsel]=[N, [Alts_Sub1, Alts_Sub2,...] ]  : s un sistema  
% (cal provar una alternativa de cada Subgraf)

sel_firstQ_inn_to_prove([[N,Qsel]],[N,Qsel]):-!.
sel_firstQ_inn_to_prove(LAltCycles,[N,Qsel]):-
	map(obtain_val_heur_inn,LAltCycles,LAltCycles1),
	sort(LAltCycles1,LAltOrd),!,  
	 % com que Heur s el primer component, ordena de menor a major segons Heur
	member22(heur(_,_,N,Qsel),LAltOrd).

% Per cada N s queda amb el cost del subgraf pitjor (Heur)
%                i la suma de costos de tots els subgrafs (SHeur)
% [N,L_Alts_Subs]=[N, [ Alts_Sub1, Alts_Sub2,...]]
obtain_val_heur_inn([N,L_Alts_Subs],heur(Heur,SHeur,N,L_Alts_Subs)):-
			% per cada Subgraf s queda amb la suma del que costa cada cicle.
	obtain_val_heur1_inn(L_Alts_Subs,Lvals),!,  % Lvals t un valor per cada subgraf
	max_list(Lvals,Heur),
	sum(Lvals,SHeur),!.

obtain_val_heur1_inn([],[]):-!.
obtain_val_heur1_inn([[_,AltsSub]|L],[Val|Lvals]):-
	obt_min_val_OR_inn(AltsSub,Val),!,
	obtain_val_heur1_inn(L,Lvals),!.


% per cada subgraf retorna el valor de l'alternativa amb valor mnim
obt_min_val_OR_inn([Alt|Alts],MinVal):-
	obt_min_val_OR_inn0(Alt,Val),
	obt_min_val_OR_inn1(Val,Alts,MinVal),!.
	
obt_min_val_OR_inn0(['a1',LAlts],MinVal):-
	obt_min_val_inn(LAlts,MinVal).
obt_min_val_OR_inn0(['a2',LAlts],MinVal):-
	obt_min_val_inn(LAlts,MinVal).
obt_min_val_OR_inn0(['b',_,LAltsSub],MinVal):-    % retorna la suma del cost mnim 
	obtain_val_eachSub_inn(LAltsSub,LVals),   % de cada subgraf.
	sum(LVals,MinValsum),
	MinVal is MinValsum-1,!.   % per prioritzar cancel.lar cicle (b) 

obtain_val_eachSub_inn([],[]):-!.
obtain_val_eachSub_inn([AltsSub|L],[Val|Lvals]):-
	obt_min_val_OR_inn0(AltsSub,Val),!,
	obtain_val_eachSub_inn(L,Lvals),!.


obt_min_val_OR_inn1(Val,[],Val):-!.
obt_min_val_OR_inn1(Val,[Alt|ORAlts],MinVal):-
	obt_min_val_OR_inn0(Alt,Val1),
	Val>Val1,!,
	obt_min_val_OR_inn1(Val1,ORAlts,MinVal),!.
obt_min_val_OR_inn1(Val,[_|ORAlts],MinVal):-
	obt_min_val_OR_inn1(Val,ORAlts,MinVal),!.


% una alternativa  [Lc,Lgr,Lge] es valora 2*|Lc|+2*|Lgr|+|Lge|
% (aix potencia les cancel.lacions de cicles perque cancel.lem un cicle(varis pairs) 
%  amb 1 constraint(pair) estricte
obt_min_val_inn([Alt|LAlts],MinVal):-
	set_val_alt_inn(Alt,Val),
	obt_min_val_inn1(Val,LAlts,MinVal),!.

set_val_alt_inn([C,Gr,Ge],Val):-
	length21(C,N1), length21(Gr,N2), length21(Ge,N3),
	Val is 2*(N1+N2)+N3,!.

obt_min_val_inn1(Val,[],Val):-!.
obt_min_val_inn1(Val,[Alt|LAlts],MinVal):-
	set_val_alt_inn(Alt,Val1),
	Val>Val1,!,
	obt_min_val_inn1(Val1,LAlts,MinVal),!.
obt_min_val_inn1(Val,[_|LAlts],MinVal):-
	obt_min_val_inn1(Val,LAlts,MinVal),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%check_inn_no_Set_setnosat(QAltern):-
%	map(head,QAltern,QAltsenum),
%	obtain_Qnoenum_inn(QAltsenum,QAlts),
%	map(sort,QAlts,QAltsOrd),
%	check_inn_no_Set_setnosat1(QAltsOrd),!.


%obtain_Qnoenum_inn([],[]):-!.
%obtain_Qnoenum_inn([Q|R],[Q1|R1]):-
%	map(selconsQ,Q,Q1),!,
%	obtain_Qnoenum_inn(R,R1),!.
	

%check_inn_no_Set_setnosat1([]):-!.
%check_inn_no_Set_setnosat1([Q|_]):-
%	nosetsatinn(Q),!,fail.
%check_inn_no_Set_setnosat1([_|R]):-
%	check_inn_no_Set_setnosat1(R),!.
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% P = 'a1' o 'a2' o [b,I,LproofSubs] i LproofSubs=[a1,a2,...] (un per subgraf)
%
proof_for_every_subgraph_inn(_,[],[],[],_):-!.
proof_for_every_subgraph_inn(C_Rules,[[Qenum,AltsSubgraph]|QAltern],[P|LP],[Res|Resultat],Help):-
	ordenar_alts_subgraph_inn(AltsSubgraph,AltsSubgraphOrd),
	member22(ORAlt,AltsSubgraphOrd), 
		% ORAlt=['a1' or 'a2',Laltscycs] o ['b',I,LSubgraphsAlts]
	proof_alt_subgraph_inn(C_Rules,Qenum,ORAlt,P,Res,Help),
	proof_for_every_subgraph_inn(C_Rules,QAltern,LP,Resultat,Help),!.

ordenar_alts_subgraph_inn([AltSubgraph],[AltSubgraph]):-!.
ordenar_alts_subgraph_inn(AltsSubgraph,AltsSubgraphOrd):-
	evaluate_alts_inn(AltsSubgraph,AltsSubgraphVal),
	sort(AltsSubgraphVal,AltsOrd),
	map(tail,AltsOrd,AltsSubgraphOrd),!.
	
evaluate_alts_inn([],[]):-!.
evaluate_alts_inn([X|Alts_b],[[Val|X]|Alts_b_Val]):-
	obt_min_val_OR_inn0(X,Val),!,
	evaluate_alts_inn(Alts_b,Alts_b_Val),!.

% ORAlt=['a1' or 'a2',Laltscycs] o ['b',I,LSubgraphsAlts]
proof_alt_subgraph_inn(C_Rules,Qenum,['b',I,LSubgraphs],['b',I,LproofSubs],L_Res,Help):-!,
	proof_for_every_subgraph_b_inn(C_Rules,Qenum,LSubgraphs,LproofSubs,L_Res,Help),!.
%K s a1 or a2
proof_alt_subgraph_inn(C_Rules,Qenum,[_,Laltscycs],K,Res,Help):-
	prove_one_set_inn(C_Rules,Qenum,Laltscycs,Res,Help),!,
	(
	 (Res=[[]|_],!, K='a2')
	 ;
	 (K='a1')
	),!.

proof_for_every_subgraph_b_inn(_,_,[],[],[],_):-!.
proof_for_every_subgraph_b_inn(C_Rules,Qenum,[[_,Laltscycs]|LSubgraphs],[K|LK],[Res|L_Res],Help):-
	prove_one_set_inn(C_Rules,Qenum,Laltscycs,Res,Help),!,
	(
	 (Res=[[]|_],!, K='a2')
	 ;
	 (K='a1')
	),!,
	proof_for_every_subgraph_b_inn(C_Rules,Qenum,LSubgraphs,LK,L_Res,Help),!.
	


%prove_one_set_inn(_,_,Qreal,_,Res,_):-
%	setsatinn(Qreal,Res),!.
prove_one_set_inn(C_Rules,Qenum,LAlts,[NewRules,Lcons,Res],Help):- 
	ord_alts_Q_inn(LAlts,LAltsOrd),
	member22(Alt,LAltsOrd),
	obtain_const_cycles_inn(Qenum,Alt,NewRules,Lcons),
	prepare_new_rules(NewRules),
	prove_set_int_rpoInn(C_Rules,NewRules,Lcons,Res,Help),!.
%	prove_set_int_rpoInn(C_Rules,NewRules,Lcons,Res,Help),
%	assertz(setsatinn(Qreal,[NewRules,Lcons,Res])),!.
%prove_one_set_inn(_,_,Qreal,_,_,_):-
%	assertz(setnosatInn(Qreal)),!,fail.


ord_alts_Q_inn([Alt],[Alt]):-!.
ord_alts_Q_inn(LAlts,LAltsOrd):-
	map(evaluateAlt,LAlts,LAltsVal),
	sort(LAltsVal,LAltsValOrd),    %ordena de menor a major [>,newRules,>=]
	map(snd2,LAltsValOrd,LAltsOrd),!.

evaluateAlt([Lc,Lgr,Lge],[[R,C,E],[Lc,Lgr,Lge]]):-
	length21(Lc,C), length21(Lgr,R), length21(Lge,E),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


prepare_new_rules(New_Rules):-
	set_new_defined(New_Rules),!.


obtain_possible_new_def([],[]):-!.
obtain_possible_new_def([Rule|NewRules],[F|D]):-
	Rule=..[_|[L,_]], L=..[F|_],
	obtain_possible_new_def(NewRules,D),!.

set_new_defined([]):-!.
set_new_defined(NewRules):-
	retractall(leftR(_)),
	retractall(noleftR(_)),
	leftR0(Def), noleftR0(Const),
	obtain_possible_new_def(NewRules,D),
	actual_leftR(Def,Const,D,Def1,Const1),
	assertz(leftR(Def1)),
	assertz(noleftR(Const1)),!.

actual_leftR(Def,Const,[],Def,Const):-!.
actual_leftR(Def,Const,[F|Rest],DefOut,ConstOut):-
	delete33([F,Ar],Const,Const1),!,
	actual_leftR([[F,Ar]|Def],Const1,Rest,DefOut,ConstOut),!.
actual_leftR(Def,Const,[_|Rest],DefOut,ConstOut):-
	actual_leftR(Def,Const,Rest,DefOut,ConstOut),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%prove_set_int_rpoInn(_,NewRules,Lcons,Res,_):-
%	satinn(NewRules,Lcons,Res),!.
prove_set_int_rpoInn(_,NewRules,Lcons,_,_):-
	nosatinn(NewRules,Lcons),!,fail.
prove_set_int_rpoInn(C_Rules,NewRules,Lcons,[IntfixOut,PrecOut,StatOut],Help):-
  	concat312(C_Rules,NewRules,C_Rules1),
	prepare_int_rpo(C_Rules1,Lcons,Prec,Stat,Intfix,Help),
	int_rpo(C_Rules1,Lcons,Intfix,IntfixOut,Prec,PrecOut,Stat,StatOut),!.
%	int_rpo(C_Rules1,Lcons,Intfix,IntfixOut,Prec,PrecOut,Stat,StatOut),
%	assertz(satinn(NewRules,Lcons,[IntfixOut,PrecOut,StatOut])),!.
prove_set_int_rpoInn(_,NewRules,Lcons,_,_):-
	check_limit_Asserts,
	assertz(nosatinn(NewRules,Lcons)),!,fail.

	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% L=[S1,S2,...] ; S1=[Qenum,LStr,LnoStr,[Op1,Op2,...]] ; 
% Op1=['a1',NumsQ,Ldcycs] or ['a2',NumsQ,Ldcycs] or 
%     ['b','i',Lconds,LcancelledCycs,LSubgrafs] with 
%	LcancelledCycs=[LcanStr,LcanNoStr]
%	LSubgrafs=[S1,S2,...] Si=['a1' or 'a2',Nums,Lstrict,LNostrict]
%
% L1=[AltS1,AltS2,...] ; 
% AltS1=['a1' or 'a2',Qenum,llistescRgrge] with 
%		llistescRgrge=[[Lc1,ll_gr1,ll_ge1],...] , or
% ['b','i',L_Alts_Subgraphs] with L_Alts_Subgraphs=[AltS11,...] (tipus 'a1','a2')
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_dif_alt_inn([N,L],[N,L1]):-
	map(obtain_dif_alt_inn_OR,L,L1),!.

obtain_dif_alt_inn_OR([Qenum,_,_,ORSubgraph],[Qenum,SubgraphAlts]):-
	map(obtain_dif_alt_inn0,ORSubgraph,SubgraphAlts),!.


obtain_dif_alt_inn0([b,I,Lconds,_,[]],[b,I,[['a1',Laltsconds]]]):-!,
	obtain_dif_alt_inn1('a1',[],[[Lconds]],Laltsconds),!.
obtain_dif_alt_inn0([b,I,Lconds,_,LSubgraphs],[b,I,LAltsSubgraphsOut]):-!, 
  % LSubgraphs=[['a1',NumsS,Lstrict,Lnostrict],['a2',...],...] - un per cada subgraf
	map(add_Lconds_to_each_subgraph,Lconds,LSubgraphs,LSubgraphs1),
	map(obtain_dif_alt_inn0,LSubgraphs1,LAltsSubgraphsOut),!.
obtain_dif_alt_inn0([K,NumsQ,Ldcycs],[K,Laltscycs]):-
	obtain_dif_alt_inn1(K,NumsQ,Ldcycs,Laltscycs),!.

% afegeix Lconds als estrictes i elimina els no estrictes
add_Lconds_to_each_subgraph(Lconds,[K,NumsQ,Ldcycs,_],[K,NumsQ,[[Lconds]|Ldcycs]]).

obtain_dif_alt_inn1(K,NumsQ,X,YOut):-
	andor_combs(X,OrX),
	dnfinn(OrX,AndOr),
%	map(del_repes,AndOr,AndOr1),
%	all_alt_LcR_nums(K,NumsQ,AndOr1,Y1),
	all_alt_LcR_nums(K,NumsQ,AndOr,Y1),
	l_concat(Y1,Y),
	simplif_alts_inn(Y,YOut),!.

andor_combs([],[]):-!.
andor_combs([X|R],[OrX|Landor]):-
	or_combs(X,OrX),!,
	andor_combs(R,Landor),!.
	
or_combs([],[]):-!.
or_combs([Lr],LcR):-!,  % Ns=[]
	conv_cR(Lr,LcR),!.
or_combs([Lr|Ns],[Ns|LcR]):-
	conv_cR(Lr,LcR),!.

conv_cR([],[]):-!.
conv_cR([S>=T|R],[cR(T>S)|R1]):-
	conv_cR(R,R1),!.
conv_cR([S>T|R],[cR(T>=S)|R1]):-
	conv_cR(R,R1),!.


dnfinn([O],O1):-!,map(list1,O,O1),!.
dnfinn([O|Landor],Lorand):-
	dnfinn(Landor,Lorand1),
	mulinn(O,Lorand1,Lorand).

mulinn([],_,[]). 
mulinn([A|R],Lorand1,Lorand):-
%	map(cons,A,Lorand1,Lorand2),
	map(addifno,A,Lorand1,Lorand2),
	mulinn(R,Lorand1,Lorand3),
	concat312(Lorand2,Lorand3,Lorand),!.


all_alt_LcR_nums(_,_,[],[]):-!.
all_alt_LcR_nums(K,NumsQ,[X|R],[X1|R1]):-
	separ_cR(X,LcR,Lnums),
        del_repes(LcR,LcR1),
	obtain_dif_alt_gr_ge1(Lnums,Lgr),!,
	l_union(Lnums,Nums),
	afegir_Lge(K,NumsQ,Nums,Lgr,Lgrge),
	map(cons,LcR1,Lgrge,X1),
	all_alt_LcR_nums(K,NumsQ,R,R1),!.

separ_cR([],[],[]):-!.
separ_cR([cR(Q)|R],[Q|R1],R2):-!,
	separ_cR(R,R1,R2),!.
separ_cR([L|R],R1,[L|R2]):-
	separ_cR(R,R1,R2),!.

afegir_Lge('a1',_,Nums,Lgr,Lgrge):-
	afegir_Lge(Nums,Lgr,Lgrge),!.
afegir_Lge('a2',NumsQ,_,Lgr,Lgrge):-
	afegir_Lge(NumsQ,Lgr,Lgrge),!.

afegir_Lge(_,[],[]):-!.
afegir_Lge(Lnums,[Lgr|LLgr],[[Lgr,Lge]|LLgrge]):-
	difer(Lnums,Lgr,Lge),!,
	afegir_Lge(Lnums,LLgr,LLgrge),!.
	
%%%%%%%%%%%%%%%

simplif_alts_inn([],[]):-!.
simplif_alts_inn([X|L],LOut):-
	is_subsumed_forSome(X,L),!,
	simplif_alts_inn(L,LOut),!.
simplif_alts_inn([X|L],[X|LOut]):-
	elim_subsumed_by(L,X,L1),
	simplif_alts_inn(L1,LOut),!.


is_subsumed_forSome(X,[Y|_]):-
	is_subsumed_c_gr_ge(X,Y),!.
is_subsumed_forSome(CC,[_|L]):-
	is_subsumed_forSome(CC,L),!.

is_subsumed_c_gr_ge([Lc,Lgr,Lge],[Lc1,Lgr1,Lge1]):-
	difer(Lc1,Lc,[]), difer(Lgr1,Lgr,[]), difer(Lge1,Lge,[]),!.

elim_subsumed_by([],_,[]):-!.
elim_subsumed_by([Y|L],X,LOut):-
	is_subsumed_c_gr_ge(Y,X),!,
	elim_subsumed_by(L,X,LOut),!.
elim_subsumed_by([Y|L],X,[Y|LOut]):-
	elim_subsumed_by(L,X,LOut),!.
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

obtain_const_cycles_inn(Qenum,[Lc,Lgr,Lge],NewRules,LCons):-
	sort(Lc,NewRules),
	obt_const_grge_inn('>',Qenum,Lgr,[],LQ1),
	obt_const_grge_inn('>=',Qenum,Lge,LQ1,LQ),
	sort(LQ,LCons),!.

obt_const_grge_inn(_,_,[],L,L):-!.
obt_const_grge_inn(Op,Q,[N|Lgr],L,LQ):-
	member22(q(N,C),Q),!,
	C=..[_|[S,T]], C1=..[Op|[S,T]],
	obt_const_grge_inn(Op,Q,Lgr,[C1|L],LQ),!.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LS=[[Qenum,Ls,Lns,OrAlt],...]
% OrAlt=[O1,O2,...], O1=['a1',Nums,Ldcyc] or ['a2',Nums,Ldcyc]  
%                 or ['b',I,Lconds,LcancelledCycs,LSubgrafs] with 
%	LcancelledCycs=[LcanStr,LcanNoStr]
%	LSubgrafs=[S1,S2,...] Si=[K,Nums,Lstrict,LNostrict] (K='a1' or 'a2')
%
% LproofSubgraphs=[PS1,PS2,...]
% PS1=[Q,Ls,Lns,a1] or [Q,Ls,Lns,a2] or [Q,Ls,Lns,[Lconds,LcancelledCycs,L1]] 
%	with L1=[[Nums,Ls,Lns,a1],[Nums,Ls,Lns,a2],...] un per subsubgraf 

obt_Lproofsubgraphs([],_,[]):-!.
obt_Lproofsubgraphs([[QN,Ls,Lns,_]|LS],[K|LProof],[[QN,Ls,Lns,K]|LproofSubgraphs]):-	 
	( K='a1' ; K='a2' ),!,
	obt_Lproofsubgraphs(LS,LProof,LproofSubgraphs),!.	 
obt_Lproofsubgraphs([[Q,Ls,Lns,LSubs]|LS],[[b,I,LproofSubs]|LProof],
			    [[Q,Ls,Lns,[Lconds,Lcancelcycs,L1]]|LproofSubgraphs]):-	 
	member21([b,I,Lconds,Lcancelcycs,LsubgraphsLconds],LSubs),!,
	obt_Lproofsubgraphs(LsubgraphsLconds,LproofSubs,L1),!,
	obt_Lproofsubgraphs(LS,LProof,LproofSubgraphs),!.	


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

is_variable(V):-variables(Vars), member212(V,Vars),!.

